@@ -2,7 +2,7 @@ package PPI::Xref;
 
 require v5.14;  # 5.10: defined-or; 5.14: package Sub { ... }
 
-our $VERSION = '0.007';
+our $VERSION = '0.008';
 
 use strict;
 use warnings;
@@ -223,8 +223,10 @@ sub __safe_dir_and_file_same_volume {
   return $self->__safe_vol_dir_file($path, 0);
 }
 
+my $CACHE_EXT = '.cache';
+
 # "shadow file" is a filename rooted into a new, "shadow", directory.
-sub __shadow_filename {
+sub __shadow_cache_filename {
     my ($self, $shadowdir, $filename) = @_;
 
     # Paranoia check.  (Either absolute or relative is fine, though.)
@@ -239,7 +241,14 @@ sub __shadow_filename {
         $filename :
         File::Spec->rel2abs($filename);
     my ($redir, $file) = $self->__safe_dir_and_file_flatten_volume($absfile);
-    return File::Spec->catfile($shadowdir, $redir, $file);
+
+    # For portable filenames, we cannot just keep on
+    # appending filename extensions with dots, and we
+    # are going to append the cache filename extension.
+    # So we mangle the .pm or .pl as _pm and _pl.
+    $file =~ s{\.(p[ml])$}{_$1};
+
+    return File::Spec->catfile($shadowdir, $redir, $file . $CACHE_EXT);
 }
 
 # The hash checksum for the file, and the mtime timestamp.
@@ -381,8 +390,8 @@ sub __write_cachefile {
 # Compose a cache filename, given an original filename.
 # The filenames are re-rooted in the cache_directory.
 sub __cache_filename {
-    my ($self, $file) = @_;
-    return if $file eq '-';
+    my ($self, $path) = @_;
+    return if $path eq '-';
 
     my $cache_directory = $self->{opt}{cache_directory};
     return unless defined $cache_directory;
@@ -392,13 +401,12 @@ sub __cache_filename {
         return;
     }
 
-    if ($file !~ /\.p[ml]$/) {
-        warn "$Sub: Unexpected filename: '$file'\n";
+    if ($path !~ /\.p[ml]$/) {
+        warn "$Sub: Unexpected filename: '$path'\n";
         return;
     }
 
-    # TODO: This will break for filesystems with only one dot allowed.
-    return $self->__shadow_filename($cache_directory, "$file.cache");
+    return $self->__shadow_cache_filename($cache_directory, $path);
 }
 
 # Deserialize from the file.
@@ -1581,7 +1589,7 @@ sub looks_like_cache_file {
 
     return 0 if $file =~ m{\.\.};
 
-    return $file =~ m{^\Q$cache_directory\E/.+.cache$};
+    return $file =~ m{^\Q$cache_directory\E/.+\Q$CACHE_EXT\E$};
 }
 
 sub cache_delete {
@@ -1593,15 +1601,17 @@ sub cache_delete {
     }
     my $delete_count = 0;
     for my $file (@_) {
-        if ($file !~ m{^/} ||
+        if (!File::Spec->file_name_is_absolute($file) ||
             $file =~ m{\.\.} ||
-            $file !~ m{\.p[ml](?:\.cache)?$}) {
+            ($file !~ m{_p[ml](?:\Q$CACHE_EXT\E)?$} &&
+             $file !~ m{.p[ml]$})) {
             # Paranoia check one.
             warn "$Sub: Skipping unexpected file: '$file'\n";
             next;
         }
         my $cache_file =
-            $file =~ /\.cache$/ ? $file : $self->__cache_filename($file);
+            $file =~ /\Q$CACHE_EXT\E$/ ?
+            $file : $self->__cache_filename($file);
         # Paranoia check two.  Both paranoia checks are needed.
         unless ($self->looks_like_cache_file($cache_file)) {
             warn "$Sub: Skipping unexpected cache file: '$cache_file'\n";
@@ -1624,7 +1634,7 @@ sub __unparse_cache_filename {
     my $cache_directory = $self->{opt}{cache_directory};
     return unless defined $cache_directory;
 
-    return unless $cache_filename =~ s{\.cache$}{};
+    return unless $cache_filename =~ s{\Q$CACHE_EXT\E$}{};
 
     my $cache_prefix_length = $self->{__cache_prefix_length};
     return unless length($cache_filename) > $cache_prefix_length;
@@ -1632,7 +1642,22 @@ sub __unparse_cache_filename {
     my $prefix = substr($cache_filename, 0, $cache_prefix_length);
     return unless $prefix =~ m{^\Q$cache_directory\E(?:/|\\)$};
 
-    return substr($cache_filename, $cache_prefix_length - 1);
+    my $path = substr($cache_filename, $cache_prefix_length - 1);
+
+    $path =~ s{_(p[ml])$}{\.$1};  # _pm -> .pm, _pl -> .pl
+
+    if ($^O eq 'MSWin32') {
+      # \c\a\b -> c:/a/b
+      $path =~ s{\\}{/}g;
+      if ($path =~ m{^/([A-Z])(/.+)}) {
+        my $volpath = "$1:$2";
+        if (-f $volpath) {
+          $path = $volpath;
+        }
+      }
+    }
+
+    return $path;
 }
 
 # Given an xref, find all the cache files under its cache directory,
@@ -1650,7 +1675,7 @@ sub find_cache_files {
 
     find(
         sub {
-            if (/\.p[ml]\.cache$/) {
+            if (/\.p[ml]\Q$CACHE_EXT\E$/) {
                 my $name = $self->__unparse_cache_filename($File::Find::name);
                 $files->{$name} = $File::Find::name;
             }
